part of nueli.core.util;

DecorConfig get $decor => DecorConfig();

typedef Decorate = dynamic Function(DecorConfig deco);

/// Decoration config
class DecorConfig extends Config<DecorConfig, Widget, Decoration> {
  Widget data;
  List<Mosaic> _hue;
  mosaic_style _mosaicStyle;
  Alignment _mosaicOrigin, _mosaicFocus;
  Alignment _alignment;
  Color _stroke;
  double _strokeSize;
  EdgeInsets _padding;
  EdgeInsets _margin;
  BoxConstraints _cons;
  Size _size;
  Clip _clip;
  RadiiConfig _radii;
  ShadowConfig _shadow;
  InkConfig _ink;

  DecorConfig() {
    _hue = [];
    _mosaicStyle = mosaic_style.linear;
    _mosaicOrigin = Alignment.centerLeft;
    _mosaicFocus = Alignment.centerRight;
    _stroke = palette.text;
    _strokeSize = 0;
    _alignment = null;
    _padding = EdgeInsets.zero;
    _margin = EdgeInsets.zero;
    _cons = null;
    _size = Size(-1, -1);
    _clip = Clip.antiAliasWithSaveLayer;
    _radii = $radii;
    _shadow = $shadow;
    _ink = $ink;
  }

  DecorConfig _config({
    List<Mosaic> hue,
    mosaic_style style,
    Alignment center,
    Alignment focus,
    Color stroke,
    double strokeSize,
    Alignment childAlignment,
    EdgeInsets padding,
    EdgeInsets margin,
    BoxConstraints cons,
    Size size,
    Clip clip,
    RadiiConfig radii,
    ShadowConfig shadow,
    InkConfig ink,
  }) {
    _hue = hue ?? _hue;
    _mosaicStyle = style ?? _mosaicStyle;
    _mosaicOrigin = center ?? _mosaicOrigin;
    _mosaicFocus = focus ?? _mosaicFocus;
    _stroke = stroke ?? _stroke;
    _strokeSize = strokeSize ?? _strokeSize;
    _alignment = childAlignment ?? _alignment;
    _padding = padding ?? _padding;
    _margin = margin ?? _margin;
    _cons = cons ?? _cons;
    _size = size ?? _size;
    _clip = clip ?? _clip;
    _radii = radii ?? _radii;
    _shadow = shadow ?? _shadow;
    _ink = ink ?? _ink;
    return this;
  }

  // input
  DecorConfig stroke(double thickness, [Color color]) => _config(stroke: color ?? palette.accent, strokeSize: thickness);

  DecorConfig radii(RadiiConfig radii) => _config(radii: radii);
  DecorConfig radii_vert(double top, double bottom) => radii($radii.vert(top, bottom)); // ignore: non_constant_identifier_names
  DecorConfig radii_horiz(double top, double bottom) => radii($radii.vert(top, bottom)); // ignore: non_constant_identifier_names
  DecorConfig radii_corners(double tl, double tr, double br, double bl) => radii($radii.corners(tl, tr, br, bl)); // ignore: non_constant_identifier_names
  DecorConfig radii_sides(double left, double right, double top, double bottom) => radii($radii.sides(left, right, top, bottom)); // ignore: non_constant_identifier_names
  DecorConfig radii_all(double value) => radii($radii.all(value)); // ignore: non_constant_identifier_names
  DecorConfig get radii_8 => radii($radii.all_8);
  DecorConfig get radii_16 => radii($radii.all_16);
  DecorConfig get radii_24 => radii($radii.all_24);
  DecorConfig get radii_52 => radii($radii.all_52);
  DecorConfig get radii_108 => radii($radii.all_108);

  DecorConfig get linear => _config(style: mosaic_style.linear);
  DecorConfig get radial => _config(style: mosaic_style.radial);
  DecorConfig get sweep => _config(style: mosaic_style.sweep);

  DecorConfig fill(List<Mosaic> hue) => _config(hue: hue);

  DecorConfig paint(Color color, [double stop]) => fill(_hue..add(Mosaic(color.value, stop)));
  DecorConfig get paint_none => paint(palette.none); // ignore: non_constant_identifier_names
  DecorConfig get paint_bg => paint(palette.background); // ignore: non_constant_identifier_names
  DecorConfig get paint_accent => paint(palette.accent); // ignore: non_constant_identifier_names
  DecorConfig get paint_primary => paint(palette.primary); // ignore: non_constant_identifier_names

  DecorConfig color(Color color) => _config(hue: [Mosaic(color.value)]);
  DecorConfig get color_none => color(palette.none); // ignore: non_constant_identifier_names
  DecorConfig get color_bg => color(palette.background); // ignore: non_constant_identifier_names
  DecorConfig get color_accent => color(palette.accent); // ignore: non_constant_identifier_names
  DecorConfig get color_primary => color(palette.primary); // ignore: non_constant_identifier_names

  DecorConfig get reverse => _config(hue: _hue.reversed.toList());

  DecorConfig orient(Alignment center, Alignment focus) => _config(center: center, focus: focus);
  DecorConfig get orient_vert => orient(Alignment.topCenter, Alignment.bottomCenter); // ignore: non_constant_identifier_names
  DecorConfig get orient_horiz => orient(Alignment.centerLeft, Alignment.centerRight); // ignore: non_constant_identifier_names
  DecorConfig get orient_diagonal => orient(Alignment.topLeft, Alignment.bottomRight); // ignore: non_constant_identifier_names
  DecorConfig get orient_xdiagonal => orient(Alignment.bottomLeft, Alignment.topRight); // ignore: non_constant_identifier_names

  DecorConfig align(Alignment alignment) => _config(childAlignment: alignment);
  DecorConfig get align_center => align(Alignment.center); // ignore: non_constant_identifier_names
  DecorConfig get align_end => align(Alignment.bottomRight); // ignore: non_constant_identifier_names
  DecorConfig get align_start => align(Alignment.topLeft); // ignore: non_constant_identifier_names
  DecorConfig get align_left => align(Alignment.centerLeft); // ignore: non_constant_identifier_names
  DecorConfig get align_right => align(Alignment.centerRight); // ignore: non_constant_identifier_names
  DecorConfig get align_top => align(Alignment.topCenter); // ignore: non_constant_identifier_names
  DecorConfig get align_bottom => align(Alignment.bottomCenter); // ignore: non_constant_identifier_names

  DecorConfig clip(Clip clip) => _config(clip: clip);
  DecorConfig get clip_none => clip(Clip.none); // ignore: non_constant_identifier_names
  DecorConfig get clip_hard => clip(Clip.hardEdge); // ignore: non_constant_identifier_names
  DecorConfig get clip_antiAlias => clip(Clip.antiAliasWithSaveLayer); // ignore: non_constant_identifier_names

  DecorConfig padding(EdgeInsets padding) => _config(padding: padding);
  DecorConfig padding_synmetric({double vert, double horiz}) => padding(EdgeInsets.symmetric(vertical: vert ?? 0, horizontal: horiz)); // ignore: non_constant_identifier_names
  DecorConfig padding_only({double left, double right, double top, double bottom}) => padding(EdgeInsets.only(left: left, right: right, top: top, bottom: bottom)); // ignore: non_constant_identifier_names
  DecorConfig padding_ltrb(double left, double top, double right, double bottom) => padding(EdgeInsets.only(left: left, right: right, top: top, bottom: bottom)); // ignore: non_constant_identifier_names
  DecorConfig padding_all(double all) => padding(EdgeInsets.all(all)); // ignore: non_constant_identifier_names
  DecorConfig get padding_8 => padding_all(8);
  DecorConfig get padding_16 => padding_all(16);
  DecorConfig get padding_zero => padding(EdgeInsets.zero); // ignore: non_constant_identifier_names

  DecorConfig margin(EdgeInsets margin) => _config(margin: margin);
  DecorConfig margin_synmetric({double vert, double horiz}) => margin(EdgeInsets.symmetric(vertical: vert ?? 0, horizontal: horiz)); // ignore: non_constant_identifier_names
  DecorConfig margin_only({double left, double right, double top, double bottom}) => margin(EdgeInsets.only(left: left, right: right, top: top, bottom: bottom)); // ignore: non_constant_identifier_names
  DecorConfig margin_ltrb(double left, double top, double right, double bottom) => margin(EdgeInsets.only(left: left, right: right, top: top, bottom: bottom)); // ignore: non_constant_identifier_names
  DecorConfig margin_all(double all) => margin(EdgeInsets.all(all)); // ignore: non_constant_identifier_names
  DecorConfig get margin_8 => margin_all(8);
  DecorConfig get margin_16 => margin_all(16);
  DecorConfig get margin_zero => margin(EdgeInsets.zero); // ignore: non_constant_identifier_names

  DecorConfig size(double w, [double h]) => _config(size: Size(w ?? 0, h ?? w ?? 0));
  DecorConfig size_h(double h) => size(_size.width, h); // ignore: non_constant_identifier_names
  DecorConfig size_w(double w) => size(w, _size.height); // ignore: non_constant_identifier_names

  DecorConfig get wrap => wrap_w.wrap_h; // ignore: non_constant_identifier_names
  DecorConfig get wrap_h => size_h(-1); // ignore: non_constant_identifier_names
  DecorConfig get wrap_w => size_w(-1); // ignore: non_constant_identifier_names

  DecorConfig get shrink => shrink_w.shrink_h; // ignore: non_constant_identifier_names
  DecorConfig get shrink_h => size_h(0); // ignore: non_constant_identifier_names
  DecorConfig get shrink_w => size_w(0); // ignore: non_constant_identifier_names

  DecorConfig get stretch => stretch_w.stretch_h;
  DecorConfig get stretch_h => size_h(double.infinity); // ignore: non_constant_identifier_names
  DecorConfig get stretch_w => size_w(double.infinity); // ignore: non_constant_identifier_names

  DecorConfig constrain(BoxConstraints cons) => _config(cons: cons);
  DecorConfig constrain_loose([double w, double h]) => constrain(BoxConstraints.loose(Size(w ?? double.infinity, h ?? w ?? double.infinity))); // ignore: non_constant_identifier_names
  DecorConfig constrain_tight([double w, double h]) => constrain(BoxConstraints.tight(Size(w ?? double.infinity, h ?? w ?? double.infinity))); // ignore: non_constant_identifier_names
  DecorConfig constrain_expand([double w, double h]) => constrain(BoxConstraints.expand(width: w, height: h)); // ignore: non_constant_identifier_names

  DecorConfig shadow(ShadowConfig shadow) => _config(shadow: shadow);
  DecorConfig raise(double elevation, [Color color]) => _config(shadow: $shadow.color(color ?? palette.shade).elevate(elevation));

  DecorConfig ink(InkConfig react) => _config(ink: react);
  DecorConfig ink_ripple(Color color, [int alpha]) => ink($ink.ripple(color, alpha)); // ignore: non_constant_identifier_names

  // operation
  @override
  DecorConfig operator *(double factor) {
    final mosaic = _hue.map((e) => e * factor).toList();

    return DecorConfig() & data
      .._config(hue: mosaic, style: _mosaicStyle, center: _mosaicOrigin * factor)
      .._config(childAlignment: _alignment == null ? null : _alignment * factor, cons: _cons)
      .._config(shadow: _shadow * factor, stroke: _stroke * factor, padding: _padding * factor)
      .._config(clip: _clip, radii: _radii * factor, size: _size * factor, margin: _margin * factor)
      .._config(strokeSize: _strokeSize * factor, ink: _ink * factor);
  }

  @override
  DecorConfig operator -(DecorConfig operand) {
    if (operand == null) return this;
    if (_mosaicStyle != operand._mosaicStyle) return this;

    final mosaic = _hue.minus(operand._hue).toList();
    final size = _size - Offset(operand._size.width, operand._size.height);

    return DecorConfig() & data
      .._config(hue: mosaic, style: _mosaicStyle, center: _mosaicOrigin - operand._mosaicOrigin)
      .._config(childAlignment: _alignment == null ? null : _alignment - operand._alignment, cons: _cons)
      .._config(shadow: _shadow - operand._shadow, stroke: _stroke - operand._stroke, padding: _padding - operand._padding)
      .._config(clip: _clip, radii: _radii - operand._radii, size: size, margin: _margin - operand._margin)
      .._config(strokeSize: _strokeSize - operand._strokeSize, ink: _ink - operand._ink);
  }

  @override
  DecorConfig operator +(DecorConfig operand) {
    if (operand == null) return this;
    if (_mosaicStyle != operand._mosaicStyle) return this;

    final mosaic = _hue.plus(operand._hue).toList();
    final size = _size + Offset(operand._size.width, operand._size.height);

    return DecorConfig() & data
      .._config(hue: mosaic, style: _mosaicStyle, center: _mosaicOrigin + operand._mosaicOrigin)
      .._config(childAlignment: _alignment == null ? null : _alignment + operand._alignment, cons: _cons)
      .._config(shadow: _shadow + operand._shadow, stroke: _stroke + operand._stroke, padding: _padding + operand._padding)
      .._config(clip: _clip, radii: _radii - operand._radii, size: size)
      .._config(margin: _margin + operand._margin, strokeSize: _strokeSize + operand._strokeSize, ink: _ink + operand._ink);
  }

  @override
  DecorConfig operator |(DecorConfig decor) {
    if (decor == null) return this;

    return DecorConfig() & data
      .._config(hue: _hue, style: _mosaicStyle)
      .._config(center: _mosaicOrigin, focus: _mosaicFocus)
      .._config(childAlignment: _alignment, cons: _cons)
      .._config(shadow: _shadow, stroke: _stroke, padding: _padding)
      .._config(clip: _clip, radii: _radii, size: _size)
      .._config(margin: _margin, strokeSize: _strokeSize, ink: _ink)
      // overlap with decor
      .._config(hue: decor._hue, style: decor._mosaicStyle)
      .._config(center: decor._mosaicOrigin, focus: decor._mosaicFocus)
      .._config(childAlignment: decor._alignment, cons: decor._cons)
      .._config(shadow: decor._shadow, stroke: decor._stroke, padding: decor._padding)
      .._config(clip: decor._clip, radii: decor._radii, size: decor._size)
      .._config(margin: decor._margin, strokeSize: decor._strokeSize, ink: decor._ink);
  }

  // output
  Gradient get gradient {
    if (_hue.isUnitary || _hue.isEmpty) return null;
    if (_mosaicStyle == mosaic_style.linear) return LinearGradient(colors: _hue, begin: _mosaicOrigin, end: _mosaicFocus, stops: _hue.stops);
    if (_mosaicStyle == mosaic_style.radial) return RadialGradient(colors: _hue, focal: _mosaicOrigin, center: _mosaicFocus, stops: _hue.stops);
    if (_mosaicStyle == mosaic_style.sweep) return SweepGradient(colors: _hue, center: _mosaicOrigin, stops: _hue.stops);
    return null;
  }

  Decoration get compile {
    return BoxDecoration(
      color: _hue.isEmpty ? palette.none : _hue.first,
      borderRadius: _radii.compile,
      gradient: gradient,
      boxShadow: _shadow != null ? (_shadow).compile : null,
      border: (_strokeSize != 0) ? Border.all(color: _stroke, width: _strokeSize) : null,
    );
  }

  Widget build({
    @deprecated Widget data,
    @deprecated Widget fallback,
    Widget child,
    Reaction reaction,
  }) {
    final result = Container(
      child: child,
      constraints: _cons,
      decoration: compile,
      width: _size.width != -1 ? _size.width : null,
      height: _size.height != -1 ? _size.height : null,
      clipBehavior: _clip,
      alignment: _alignment ?? (_size.width == double.infinity ? Alignment.center : null),
      padding: _padding,
      margin: _margin,
    );
    if (reaction == null && _ink.data == null) return result;
    return Material(color: palette.none, child: (_ink).radii(_radii).build(child: result, data: reaction));
  }
}

enum mosaic_style { linear, radial, sweep }

class Mosaic extends Color with Subtract<Mosaic>, Add<Mosaic> {
  final int value;
  final double stop;

  Mosaic(this.value, [double stop])
      : stop = stop ?? 1,
        super(value);

  Mosaic operator *(double factor) {
    return Mosaic((Color(value) * factor).value, stop * factor);
  }

  Mosaic operator -(Mosaic operand) {
    return Mosaic((Color(value) - Color(operand.value)).value, stop - operand.stop);
  }

  Mosaic operator +(Mosaic operand) {
    return Mosaic((Color(value) + Color(operand.value)).value, stop + operand.stop);
  }
}

extension Hue on List<Mosaic> {
  double get _totalStop => fold<double>(0, (previousValue, mosaic) => previousValue + mosaic.stop);
  List<double> get stops {
    double former = 0;
    List<double> stops = [];
    for (int i = 0; i < length; i++) {
      Mosaic mosaic = this[i];
      stops.add((former + mosaic.stop / 2 * i / length) / _totalStop);
      former += mosaic.stop;
    }
    return stops;
  }
}

extension ColorConfig on Color {
  Mosaic operator &(double stop) => Mosaic(value, stop);

  Color operator *(double factor) => Color.fromRGBO((red * factor).round(), (green * factor).round(), (blue * factor).round(), opacity * factor);
  Color operator -(Color color) => Color.fromRGBO(red - color.red, green - color.green, blue - color.blue, opacity - color.opacity);
  Color operator +(Color color) => Color.fromRGBO(red + color.red, green + color.green, blue + color.blue, opacity + color.opacity);
}
