/// embed [Bucket]
part of nueli.core.util;

IconConfig get $icon => IconConfig();

/// Icon config
class IconConfig extends Config<IconConfig, IconData, IconData> {
  IconData data;
  bool _useColor;
  BlendMode _blendMode;
  Color _color;
  Size _size;
  Alignment _align;
  BoxFit _boxFit;

  IconConfig() {
    data = null;
    _color = palette.text;
    _size = Size.square(24);
    _blendMode = BlendMode.srcIn;
    _useColor = true;
    _align = Alignment.center;
    _boxFit = BoxFit.contain;
  }

  IconConfig _config({
    Color color,
    Size size,
    BlendMode blendMode,
    bool filter,
    Alignment align,
    BoxFit boxFit,
  }) {
    _color = color ?? _color;
    _size = size ?? _size;
    _blendMode = blendMode ?? _blendMode;
    _useColor = filter ?? _useColor;
    _align = align ?? _align;
    _boxFit = boxFit ?? _boxFit;
    return this;
  }

  // input
  IconConfig color(Color color) => _config(color: color);
  IconConfig get color_bg => color(palette.background); // ignore: non_constant_identifier_names
  IconConfig get color_text => color(palette.text); // ignore: non_constant_identifier_names
  IconConfig get color_hint => color(palette.text.withAlpha(150)); // ignore: non_constant_identifier_names
  IconConfig get color_clear => _config(filter: false); // ignore: non_constant_identifier_names

  IconConfig filter(bool filter) => _config(filter: filter);

  IconConfig align(Alignment align) => _config(align: align);
  IconConfig get align_start => align(Alignment.topLeft); // ignore: non_constant_identifier_names
  IconConfig get align_center => align(Alignment.center); // ignore: non_constant_identifier_names
  IconConfig get align_end => align(Alignment.bottomRight); // ignore: non_constant_identifier_names

  IconConfig fit(BoxFit fit) => _config(boxFit: fit);
  IconConfig get fit_contain => fit(BoxFit.contain); // ignore: non_constant_identifier_names
  IconConfig get fit_fill => fit(BoxFit.fill); // ignore: non_constant_identifier_names
  IconConfig get fit_cover => fit(BoxFit.cover); // ignore: non_constant_identifier_names
  IconConfig get fit_center => fit(BoxFit.scaleDown); // ignore: non_constant_identifier_names
  IconConfig get fit_height => fit(BoxFit.fitHeight); // ignore: non_constant_identifier_names
  IconConfig get fit_width => fit(BoxFit.fitWidth); // ignore: non_constant_identifier_names
  IconConfig get fit_none => fit(BoxFit.none); // ignore: non_constant_identifier_names

  IconConfig size(double w, [double h]) => _config(size: Size(w ?? 0, h ?? w ?? 0));
  IconConfig size_w(double w) => size(w, _size.height); // ignore: non_constant_identifier_names
  IconConfig size_h(double h) => size(_size.width, h); // ignore: non_constant_identifier_names
  IconConfig get size_24 => size(24);
  IconConfig get size_16 => size(16);
  IconConfig get size_36 => size(36);
  IconConfig get size_48 => size(48);
  IconConfig get size_72 => size(72);
  IconConfig get size_100 => size(100);
  IconConfig get size_120 => size(120);

  IconConfig get wrap_w => size_w(0); // ignore: non_constant_identifier_names
  IconConfig get wrap_h => size_h(0); // ignore: non_constant_identifier_names
  IconConfig get wrap_both => size(0, 0); // ignore: non_constant_identifier_names

  IconConfig get stretch_w => size_w(double.infinity); // ignore: non_constant_identifier_names
  IconConfig get stretch_h => size_h(double.infinity); // ignore: non_constant_identifier_names
  IconConfig get stretch_both => size(double.infinity, double.infinity); // ignore: non_constant_identifier_names

  // operation
  @override
  IconConfig operator *(double factor) {
    return IconConfig() & data
      .._config(align: _align * factor, color: Color((_color.value * factor).round()), size: _size * factor) // break
      .._config(boxFit: _boxFit, filter: _useColor, blendMode: _blendMode);
  }

  @override
  IconConfig operator -(IconConfig operand) {
    return IconConfig() & data
      .._config(align: _align - operand._align, color: _color, size: _size - operand._size) // break
      .._config(boxFit: _boxFit, filter: _useColor, blendMode: _blendMode);
  }

  @override
  IconConfig operator +(IconConfig operand) {
    final size = _size + Offset(operand._size.width, operand._size.height);

    return IconConfig()
      .._config(align: _align + operand._align, color: _color, size: size) // break
      .._config(boxFit: _boxFit, filter: _useColor, blendMode: _blendMode);
  }

  @override
  IconConfig operator |(IconConfig ico) {
    return IconConfig() & data
      .._config(align: _align, color: _color, size: _size, boxFit: _boxFit)
      .._config(filter: _useColor, blendMode: _blendMode)
      // overlap with ico
      .._config(align: ico._align, color: ico._color, size: ico._size, boxFit: ico._boxFit)
      .._config(filter: ico._useColor, blendMode: ico._blendMode);
  }

  // output
  IconData get compile => data;

  @override
  Widget build({IconData data, IconData fallback = CupertinoIcons.exclamationmark_circle}) => get(data ?? this.data ?? fallback);

  Widget get(IconData data) => Icon(data, color: _color, size: _size.shortestSide);

  Widget load(String path) => Image.asset(
        path,
        fit: _boxFit,
        alignment: _align,
        color: _useColor ? _color : null,
        colorBlendMode: _useColor ? _blendMode : null,
        height: (_size.height) == 0 ? null : (_size.height),
        width: (_size.width) == 0 ? null : (_size.width),
      );

  Widget fetch(String src) {
    Widget loader(ImageChunkEvent chunk) {
      final loadValue = chunk != null ? (chunk.expectedTotalBytes != null ? chunk.cumulativeBytesLoaded / chunk.expectedTotalBytes : null) : null;
      return Center(
        child: Container(
          height: 36,
          width: 36,
          padding: const EdgeInsets.all(8.0),
          child: CircularProgressIndicator(
            strokeWidth: 2,
            backgroundColor: palette.background,
            value: loadValue,
          ),
        ),
      );
    }

    Widget error(StackTrace trace) {
      final errType = _size.width > 100 && _size.height > 50;
      final errIco = $icon.color(palette.error) & CupertinoIcons.exclamationmark_circle;
      final errText = $text.align_center.color((_color ?? palette.text).withAlpha(100)).build(data: 'An error occurred');
      return Center(child: Padding(padding: const EdgeInsets.all(8.0), child: errType ? errText : errIco));
    }

    return Image.network(
      src,
      color: (_useColor ? _color : null),
      fit: _boxFit,
      alignment: _align,
      colorBlendMode: _blendMode,
      height: (_size.height) == 0 ? null : (_size.height),
      width: (_size.width) == 0 ? null : (_size.width),
      errorBuilder: (c, o, s) => error(s),
      loadingBuilder: (c, child, chunk) => chunk == null ? child : loader(chunk),
    );
  }
}

extension LibraryIcons on IconConfig {
  Widget get flash => get(Icons.flash_on);
  Widget get flashOff => get(Icons.flash_off);
  Widget get favourite => get(CupertinoIcons.suit_heart);
  Widget get favouriteFill => get(CupertinoIcons.suit_heart_fill);
  Widget get check => get(CupertinoIcons.app);
  Widget get checkFill => get(CupertinoIcons.checkmark_square_fill);
}
