part of nueli.core.util;

InkConfig get $ink => InkConfig();

class InkConfig extends Config<InkConfig, Reaction, InkConfig> {
  Reaction data;
  Color _splash;
  Color _hover;
  Color _focus;
  Color _overlay;
  Color _highlight;
  double _size;
  bool _response;
  RadiiConfig _radii;

  InkConfig() {
    _splash = palette.accent.withAlpha(palette.tint.alpha);
    _hover = palette.accent.withAlpha(palette.tint.alpha ~/ 2);
    _focus = palette.accent.withAlpha(palette.tint.alpha ~/ 5);
    _overlay = palette.accent.withAlpha(palette.tint.alpha ~/ 2);
    _highlight = palette.accent.withAlpha(palette.tint.alpha ~/ 3);
    _size = 200;
    _response = false;
    _radii = $radii.zero;
  }

  InkConfig _config({
    Color splash,
    Color hover,
    Color focus,
    Color overlay,
    Color highlight,
    double radius,
    bool response,
    RadiiConfig radii,
  }) {
    _splash = splash ?? _splash;
    _hover = hover ?? _hover;
    _focus = focus ?? _focus;
    _overlay = overlay ?? _overlay;
    _highlight = highlight ?? _highlight;
    _size = radius ?? _size;
    _response = response ?? _response;
    _radii = radii ?? _radii;
    return this;
  }

  // input
  InkConfig well(double size) => _config(response: false, radius: size);
  InkConfig response(double size) => _config(response: true, radius: size);

  InkConfig ripple(Color color, [int alpha]) => splash(color, alpha).hover(color, alpha).overlay(color, alpha);
  InkConfig splash(Color color, [int alpha]) => _config(splash: color.withAlpha(alpha ?? palette.tint.alpha));
  InkConfig hover(Color color, [int alpha]) => _config(hover: color.withAlpha(alpha ?? palette.tint.alpha ~/ 2));
  InkConfig focus(Color color, [int alpha]) => _config(focus: color.withAlpha(alpha ?? palette.tint.alpha ~/ 5));
  InkConfig overlay(Color color, [int alpha]) => _config(overlay: color.withAlpha(alpha ?? palette.tint.alpha ~/ 2));
  InkConfig highlight(Color color, [int alpha]) => _config(highlight: color.withAlpha(alpha ?? palette.tint.alpha ~/ 3));

  InkConfig radii(RadiiConfig radii) => _config(radii: radii);
  InkConfig radii_vert(double top, double bottom) => radii($radii.vert(top, bottom)); // ignore: non_constant_identifier_names
  InkConfig radii_horiz(double top, double bottom) => radii($radii.vert(top, bottom)); // ignore: non_constant_identifier_names
  InkConfig radii_corners(double tl, double tr, double br, double bl) => radii($radii.corners(tl, tr, br, bl)); // ignore: non_constant_identifier_names
  InkConfig radii_sides(double left, double right, double top, double bottom) => radii($radii.sides(left, right, top, bottom)); // ignore: non_constant_identifier_names
  InkConfig radii_all(double value) => radii($radii.all(value)); // ignore: non_constant_identifier_names
  InkConfig get radii_8 => radii($radii.all_8);
  InkConfig get radii_16 => radii($radii.all_16);
  InkConfig get radii_24 => radii($radii.all_24);
  InkConfig get radii_52 => radii($radii.all_52);
  InkConfig get radii_108 => radii($radii.all_108);

  @override
  InkConfig operator *(double factor) {
    return InkConfig() & data
      .._config(splash: Color((_splash.value * factor).round()))
      .._config(hover: Color((_hover.value * factor).round()))
      .._config(focus: Color((_focus.value * factor).round()))
      .._config(overlay: Color((_overlay.value * factor).round()))
      .._config(highlight: Color((_highlight.value * factor).round()))
      .._config(radius: _size * factor, response: _response)
      .._config(radii: _radii * factor);
  }

  // operation
  @override
  InkConfig operator -(InkConfig operand) {
    return InkConfig() & data
      .._config(splash: Color(_splash.value - operand._splash.value))
      .._config(hover: Color(_hover.value - operand._hover.value))
      .._config(focus: Color(_focus.value - operand._focus.value))
      .._config(overlay: Color(_overlay.value - operand._overlay.value))
      .._config(highlight: Color(_highlight.value - operand._highlight.value))
      .._config(radius: _size - operand._size, response: _response)
      .._config(radii: _radii - operand._radii);
  }

  @override
  InkConfig operator +(InkConfig operand) {
    return InkConfig() & data
      .._config(splash: Color(_splash.value + operand._splash.value))
      .._config(hover: Color(_hover.value + operand._hover.value))
      .._config(focus: Color(_focus.value + operand._focus.value))
      .._config(overlay: Color(_overlay.value + operand._overlay.value))
      .._config(highlight: Color(_highlight.value + operand._highlight.value))
      .._config(radius: _size + operand._size, response: _response)
      .._config(radii: _radii + operand._radii);
  }

  @override
  InkConfig operator |(InkConfig react) {
    return InkConfig() & data
      .._config(splash: _splash, hover: _hover, focus: _focus, overlay: _overlay, highlight: _highlight)
      .._config(radius: _size, response: _response, radii: _radii)
      // overlap with matrix
      .._config(splash: react._splash, hover: react._hover, focus: react._focus, overlay: react._overlay)
      .._config(highlight: react._highlight, radius: react._size, response: react._response, radii: react._radii);
  }

  // output
  @deprecated
  InkConfig get compile => this;

  Widget build({Reaction data, Reaction fallback, Widget child}) {
    this.data = data ?? this.data ?? fallback ?? () {};
    return _response
        ? InkResponse(
            child: child,
            highlightColor: Color((_overlay.value).round()),
            overlayColor: MaterialStateProperty.all<Color>(Color((_overlay.value).round())),
            splashColor: Color((_splash.value).round()),
            hoverColor: Color((_hover.value).round()),
            focusColor: Color((_focus.value).round()),
            radius: max(Material.defaultSplashRadius, _size * 0.7),
            onTap: this.data.onTap,
            onLongPress: this.data.onLongPress,
          )
        : InkWell(
            child: child,
            borderRadius: _radii.compile,
            splashColor: Color((_splash.value).round()),
            hoverColor: Color((_hover.value).round()),
            focusColor: Color((_focus.value).round()),
            overlayColor: MaterialStateProperty.all<Color>(Color((_overlay.value).round())),
            highlightColor: Color((_highlight.value).round()),
            radius: _size,
            onTap: this.data.onTap,
            onLongPress: this.data.onLongPress,
          );
  }
}

class Reaction {
  final Function() onTap;
  final Function() onLongPress;
  Reaction({this.onTap, this.onLongPress});
}

class ReactData {}
