part of nueli.core.util;

MatrixConfig get $matrix => MatrixConfig();

class MatrixConfig extends Config<MatrixConfig, Matrix4, Matrix4> {
  Matrix4 data;
  Alignment _alignment;
  List<_TransformData> _transforms;

  MatrixConfig() {
    _transforms = [];
    _alignment = Alignment.center;
  }

  MatrixConfig _config({
    Alignment alignment,
    List<_TransformData> transforms,
  }) {
    _alignment = alignment ?? _alignment;
    _transforms = _transforms..addAll(transforms);
    return this;
  }

  // input
  MatrixConfig align(Alignment alignment) => _config(alignment: alignment);

  MatrixConfig skew({double x = 0, double y = 0, double z = 0}) => (skew_x(x)..skew_y(y))..rotate(z);
  MatrixConfig skew_x(double x) => _config(transforms: [_TransformData(x, (t, d) => t..rotateX(d))]); // ignore: non_constant_identifier_names
  MatrixConfig skew_y(double y) => _config(transforms: [_TransformData(y, (t, d) => t..rotateY(d))]); // ignore: non_constant_identifier_names
  MatrixConfig rotate(double z) => _config(transforms: [_TransformData(z, (t, d) => t..rotateZ(d))]); // ignore: non_constant_identifier_names

  MatrixConfig zoom(double zoom) => _config(transforms: [_TransformData(zoom, (t, d) => t..rotateZ(d))]);

  MatrixConfig pan({double x = 0, double y = 0, double z = 0}) => (pan_x(x)..pan_y(y))..pan_z(z);
  MatrixConfig pan_x(double x) => _config(transforms: [_TransformData(x, (t, d) => t..translate(d))]); // ignore: non_constant_identifier_names
  MatrixConfig pan_y(double y) => _config(transforms: [_TransformData(y, (t, d) => t..translate(0, d))]); // ignore: non_constant_identifier_names
  MatrixConfig pan_z(double z) => _config(transforms: [_TransformData(z, (t, d) => t..translate(0, 0, d))]); // ignore: non_constant_identifier_names

  MatrixConfig shift({double x = 0, double y = 0, double z = 0}) => (shift_x(x)..shift_y(y))..shift_z(z);
  MatrixConfig shift_x(double x) => _config(transforms: [_TransformData(x, (t, d) => t..leftTranslate(d))]); // ignore: non_constant_identifier_names
  MatrixConfig shift_y(double y) => _config(transforms: [_TransformData(y, (t, d) => t..leftTranslate(0, d))]); // ignore: non_constant_identifier_names
  MatrixConfig shift_z(double z) => _config(transforms: [_TransformData(z, (t, d) => t..leftTranslate(0, 0, d))]); // ignore: non_constant_identifier_names

  MatrixConfig scale({double x = 0, double y = 0, double z = 0}) => (scale_x(x)..scale_y(y))..scale_z(z);
  MatrixConfig scale_x(double x) => _config(transforms: [_TransformData(x, (t, d) => t..scale(d))]); // ignore: non_constant_identifier_names
  MatrixConfig scale_y(double y) => _config(transforms: [_TransformData(y, (t, d) => t..scale(0, d))]); // ignore: non_constant_identifier_names
  MatrixConfig scale_z(double z) => _config(transforms: [_TransformData(z, (t, d) => t..scale(0, 0, d))]); // ignore: non_constant_identifier_names

  // operation
  @override
  MatrixConfig operator *(double factor) {
    List<_TransformData> transforms = _transforms.map((e) => e * factor);
    return MatrixConfig() & data
      .._config(alignment: _alignment * factor, transforms: transforms);
  }

  @override
  MatrixConfig operator -(MatrixConfig operand) {
    List<_TransformData> transforms = _transforms.minus(operand._transforms);
    return MatrixConfig() & data
      .._config(alignment: _alignment - operand._alignment, transforms: transforms);
  }

  @override
  MatrixConfig operator +(MatrixConfig operand) {
    List<_TransformData> transforms = _transforms.plus(operand._transforms);
    return MatrixConfig() & data
      .._config(alignment: _alignment + operand._alignment, transforms: transforms);
  }

  @override
  MatrixConfig operator |(MatrixConfig matrix) {
    return MatrixConfig() & data
      .._config(alignment: _alignment, transforms: _transforms)
      // overlap with matrix
      .._config(alignment: matrix._alignment, transforms: matrix._transforms);
  }

  // output
  Matrix4 get compile {
    Matrix4 mat = data;
    for (_TransformData transform in _transforms) mat = transform(mat);
    return mat;
  }

  Widget build({@deprecated Matrix4 data, @deprecated Matrix4 fallback, Widget child}) {
    this.data = data ?? this.data ?? fallback ?? Matrix4.identity()
      ..setEntry(3, 2, 0.001);
    _alignment = _alignment ?? Alignment.center;
    return Transform(child: child, alignment: _alignment, transform: compile);
  }
}

class _TransformData with Subtract<_TransformData>, Add<_TransformData> {
  double value;
  Matrix4 Function(Matrix4, double) transform;
  _TransformData(this.value, this.transform);
  Matrix4 call(Matrix4 matrix) => transform(matrix, value);

  _TransformData operator -(_TransformData element) {
    return _TransformData(value - element.value, (mat, d) => transform(element.transform(mat, d), d));
  }

  @override
  _TransformData operator +(_TransformData element) {
    return _TransformData(value + element.value, (mat, d) => transform(element.transform(mat, d), d));
  }

  _TransformData operator *(double factor) {
    return _TransformData(value + factor, (mat, d) => transform(mat, d));
  }
}
