part of nueli.core.util;

RadiiConfig get $radii => RadiiConfig();

class RadiiConfig extends Config<RadiiConfig, Widget, BorderRadius> {
  Widget data;
  Radius _topLeft;
  Radius _topRight;
  Radius _bottomRight;
  Radius _bottomLeft;

  RadiiConfig() {
    _topLeft = Radius.zero;
    _topRight = Radius.zero;
    _bottomRight = Radius.zero;
    _bottomLeft = Radius.zero;
  }

  RadiiConfig _config({
    Radius topLeft,
    Radius topRight,
    Radius bottomRight,
    Radius bottomLeft,
  }) {
    _topLeft = topLeft ?? _topLeft;
    _topRight = topRight ?? _topRight;
    _bottomRight = bottomRight ?? _bottomRight;
    _bottomLeft = bottomLeft ?? _bottomLeft;
    return this;
  }

  // input
  RadiiConfig top_left(double x, [double y]) => _config(topLeft: Radius.elliptical(x, y ?? x)); // ignore: non_constant_identifier_names
  RadiiConfig top_right(double x, [double y]) => _config(topRight: Radius.elliptical(x, y ?? x)); // ignore: non_constant_identifier_names
  RadiiConfig bottom_right(double x, [double y]) => _config(bottomRight: Radius.elliptical(x, y ?? x)); // ignore: non_constant_identifier_names
  RadiiConfig bottom_left(double x, [double y]) => _config(bottomLeft: Radius.elliptical(x, y ?? x)); // ignore: non_constant_identifier_names

  RadiiConfig left(double value) => top_left(_topLeft.x, value).bottom_left(_bottomLeft.x, value);
  RadiiConfig right(double value) => top_right(_topRight.x, value).bottom_right(_bottomRight.x, value);
  RadiiConfig top(double value) => top_left(value, _topLeft.y).top_right(value, _topRight.y);
  RadiiConfig bottom(double value) => bottom_left(value, _bottomLeft.y).bottom_right(value, _bottomRight.y);

  RadiiConfig vert(double top, double bottom) => top_left(top).top_right(top).bottom_left(bottom).bottom_right(bottom);
  RadiiConfig horiz(double left, double right) => top_left(left).top_right(right).bottom_left(left).bottom_right(right);

  RadiiConfig corners(double tl, double tr, double br, double bl) => top_left(tl).top_right(tr).bottom_right(br).bottom_left(bl); // ignore: non_constant_identifier_names
  RadiiConfig sides(double left, double right, double top, double bottom) => this.left(left).right(right).top(top).bottom(bottom); // ignore: non_constant_identifier_names
  
  RadiiConfig all(double value) => left(value).right(value).top(value).bottom(value);
  RadiiConfig get all_8 => all(8);
  RadiiConfig get all_16 => all(16);
  RadiiConfig get all_24 => all(24);
  RadiiConfig get all_52 => all(52);
  RadiiConfig get all_108 => all(108);
  RadiiConfig get zero => all(0);

  // operation
  @override
  RadiiConfig operator *(double factor) {
    return RadiiConfig() & data
      .._config(topLeft: _topLeft * factor, topRight: _topRight * factor)
      .._config(bottomRight: _bottomRight * factor, bottomLeft: _bottomLeft * factor);
  }

  @override
  RadiiConfig operator -(RadiiConfig operand) {
    return RadiiConfig() & data
      .._config(topLeft: _topLeft - operand._topLeft, topRight: _topRight - operand._topRight)
      .._config(bottomRight: _bottomRight - operand._bottomRight, bottomLeft: _bottomLeft - operand._bottomLeft);
  }

  @override
  RadiiConfig operator +(RadiiConfig operand) {
    return RadiiConfig() & data
      .._config(topLeft: _topLeft + operand._topLeft, topRight: _topRight + operand._topRight)
      .._config(bottomRight: _bottomRight + operand._bottomRight, bottomLeft: _bottomLeft + operand._bottomLeft);
  }

  @override
  RadiiConfig operator |(RadiiConfig radii) {
    return RadiiConfig() & data
      .._config(topLeft: _topLeft, topRight: _topRight)
      .._config(bottomRight: _bottomRight, bottomLeft: _bottomLeft)
      // overlap with matrix
      .._config(topLeft: radii._topLeft, topRight: radii._topRight)
      .._config(bottomRight: radii._bottomRight, bottomLeft: radii._bottomLeft);
  }

  // output
  BorderRadius get compile {
    return BorderRadius.only(
      topLeft: _topLeft,
      topRight: _topRight,
      bottomRight: _bottomRight,
      bottomLeft: _bottomLeft,
    );
  }

  Widget build({@deprecated Widget data, @deprecated Widget fallback, Widget child}) {
    return Container(
      child: child,
      decoration: BoxDecoration(borderRadius: compile),
      clipBehavior: Clip.antiAliasWithSaveLayer,
    );
  }
}
