part of nueli.core.util;

ShadowConfig get $shadow => ShadowConfig();

class ShadowConfig extends Config<ShadowConfig, Widget, List<BoxShadow>> {
  List<ShadowConfig> _children;
  Widget data;
  Color _color;
  Offset _offset;
  double _spread;
  double _blur;

  ShadowConfig([this._children = const []]) {
    data = null;
    _color = palette.none;
    _offset = Offset.zero;
    _spread = 0;
    _blur = 0;
  }

  ShadowConfig _config({
    Widget child,
    Color color,
    Offset offset,
    double spread,
    double blur,
  }) {
    data = child ?? data;
    _color = color ?? _color;
    _offset = offset ?? _offset;
    _spread = spread ?? _spread;
    _blur = blur ?? _blur;
    return this;
  }

  // input
  ShadowConfig color(Color color) => _config(color: color);
  ShadowConfig grow(double x, [double y]) => _config(offset: Offset(x ?? 0, y ?? x ?? 0));
  ShadowConfig blur(double blur) => _config(blur: blur); // ignore: non_constant_identifier_names
  ShadowConfig spread(double spread) => _config(spread: spread); // ignore: non_constant_identifier_names
  ShadowConfig glow(double glow) => _config(offset: Offset.zero, blur: glow);

  ShadowConfig elevate(double elevation) => _config(blur: elevation, spread: -0.5, offset: Offset(elevation / 8, elevation / 8));
  ShadowConfig float(double float) => _config(blur: float / 6, spread: -0.5, offset: Offset(float, float));

  ShadowConfig get save => ShadowConfig(_children..add(this.._children = []));

  // operation
  @override
  ShadowConfig operator *(double factor) {
    return ShadowConfig() & data
      .._children.map((e) => e * factor).toList()
      .._config(child: data, color: _color * factor, offset: _offset * factor) // break
      .._config(spread: _spread * factor, blur: _blur * factor);
  }

  @override
  ShadowConfig operator -(ShadowConfig operand) {
    final color = Color(_color.value - operand._color.value);

    return ShadowConfig() & data
      .._children.map((e) => e - operand).toList()
      .._config(child: data, color: color, offset: _offset - operand._offset) // break
      .._config(spread: _spread - operand._spread, blur: _blur - operand._blur);
  }

  @override
  ShadowConfig operator +(ShadowConfig operand) {
    final color = Color(_color.value + operand._color.value);

    return ShadowConfig() & data
      .._config(color: color, offset: _offset + operand._offset) // break
      .._config(spread: _spread + operand._spread, blur: _blur + operand._blur);
  }

  @override
  ShadowConfig operator |(ShadowConfig shadow) {
    return ShadowConfig() & data
      .._config(color: _color, blur: _blur, spread: _spread, offset: _offset)
      // overlap with matrix
      .._config(color: shadow._color, blur: shadow._blur, spread: shadow._spread, offset: shadow._offset);
  }

  // output
  List<BoxShadow> get compile => (_children.toList()..add(this))
      .map((e) => BoxShadow(
            color: _color,
            offset: e._offset,
            blurRadius: e._blur,
            spreadRadius: e._spread,
          ))
      .toList();

  Widget build({@deprecated Widget data, @deprecated Widget fallback, Widget child, bool displayInFront = false}) {
    data = child ?? data ?? SizedBox();
    final posit = displayInFront ? DecorationPosition.foreground : DecorationPosition.background;
    return DecoratedBox(decoration: BoxDecoration(boxShadow: compile), child: data, position: posit);
  }
}
