part of nueli.core.util;

TextConfig get $text => TextConfig();

/// Text config
class TextConfig extends Config<TextConfig, String, TextStyle> {
  String data;
  Color _color, _bcolor;
  double _size;
  bool _underline, _strike, _overline, _italic, _fill;
  int _maxLines;
  FontWeight _weight;
  TextAlign _align;
  ShadowConfig _shadow;
  InkConfig _ink;

  TextConfig _config({
    Color color,
    Color background,
    double size,
    bool underline,
    bool strike,
    bool overline,
    bool italic,
    FontWeight weight,
    TextAlign align,
    bool fill,
    int maxLines,
    ShadowConfig shadow,
    InkConfig ink,
  }) {
    _color = color ?? _color;
    _bcolor = background ?? _bcolor;
    _size = size ?? _size;
    _shadow = shadow ?? _shadow;
    _underline = underline ?? _underline;
    _overline = overline ?? _overline;
    _strike = strike ?? _strike;
    _italic = italic ?? _italic;
    _weight = weight ?? _weight;
    _align = align ?? _align;
    _fill = fill ?? _fill;
    _maxLines = maxLines ?? _maxLines;
    _ink = ink ?? _ink;
    return this;
  }

  // input
  TextConfig color(Color color) => _config(color: color);
  TextConfig get color_bg => _config(color: palette.background); // ignore: non_constant_identifier_names
  TextConfig get color_text => _config(color: palette.text); // ignore: non_constant_identifier_names
  TextConfig get color_hint => _config(color: palette.text.withAlpha(150)); // ignore: non_constant_identifier_names

  TextConfig line(int max) => _config(maxLines: max <= 0 ? null : max); // ignore: non_constant_identifier_names
  TextConfig get line_mono => line(1); // ignore: non_constant_identifier_names
  TextConfig get line_unlimited => line(-1); // ignore: non_constant_identifier_names

  TextConfig get strike => _config(strike: true);
  TextConfig get overline => _config(overline: true);
  TextConfig get underline => _config(underline: true);

  TextConfig get style_italic => _config(italic: true); // ignore: non_constant_identifier_names
  TextConfig get style_bold => weight(FontWeight.bold); // ignore: non_constant_identifier_names

  TextConfig weight(FontWeight weight) => _config(weight: weight);
  TextConfig get weight_bold => weight(FontWeight.bold); // ignore: non_constant_identifier_names
  TextConfig get weight_900 => weight(FontWeight.w900);
  TextConfig get weight_700 => weight(FontWeight.w700);
  TextConfig get weight_500 => weight(FontWeight.w500);
  TextConfig get weight_200 => weight(FontWeight.w300);
  TextConfig get weight_100 => weight(FontWeight.w100);

  TextConfig size(double size) => _config(size: size);
  TextConfig get size_10 => size(10);
  TextConfig get size_12 => size(12);
  TextConfig get size_14 => size(14);
  TextConfig get size_16 => size(16);
  TextConfig get size_18 => size(18);
  TextConfig get size_24 => size(24);

  TextConfig shadow(ShadowConfig shadow) => _config(shadow: shadow);
  TextConfig elevate({Color color, double blur}) => shadow($shadow.color(color).blur(blur));

  TextConfig align(TextAlign textAlign) => _config(align: textAlign);
  TextConfig get align_center => align(TextAlign.center); // ignore: non_constant_identifier_names
  TextConfig get align_end => align(TextAlign.end); // ignore: non_constant_identifier_names
  TextConfig get align_justify => align(TextAlign.justify); // ignore: non_constant_identifier_names
  TextConfig get align_start => align(TextAlign.start); // ignore: non_constant_identifier_names

  TextConfig get stretch => _config(fill: true);
  TextConfig get shrink => _config(fill: false);

  TextConfig ink(InkConfig react) => _config(ink: react);
  TextConfig ink_ripple(Color color, [int alpha]) => ink($ink.ripple(color, alpha)); // ignore: non_constant_identifier_names

  // operation
  @override
  TextConfig operator *(double factor) {
    final color = Color((_color.value * factor).round());
    final bcolor = Color((_bcolor.value * factor).round());
    return TextConfig() & data
      .._config(color: color, background: bcolor, size: _size * factor)
      .._config(underline: _underline, strike: _strike, overline: _overline, italic: _italic)
      .._config(weight: _weight, shadow: _shadow * factor, ink: _ink * factor);
  }

  @override
  TextConfig operator -(TextConfig operand) {
    final color = Color(_color.value - operand._color.value);
    final bcolor = Color(_bcolor.value - operand._bcolor.value);
    return TextConfig() & data
      .._config(color: color, background: bcolor, size: _size - operand._size)
      .._config(underline: _underline, strike: _strike, overline: _overline, italic: _italic)
      .._config(weight: _weight, shadow: _shadow - operand._shadow, ink: _ink - operand._ink);
  }

  @override
  TextConfig operator +(TextConfig operand) {
    final color = Color(_color.value + operand._color.value);
    final bcolor = Color(_bcolor.value + operand._bcolor.value);
    return TextConfig() & data
      .._config(color: color, background: bcolor, size: _size + operand._size)
      .._config(underline: _underline, strike: _strike, overline: _overline, italic: _italic)
      .._config(weight: _weight, shadow: _shadow + operand._shadow, ink: _ink + operand._ink);
  }

  TextConfig operator |(TextConfig text) {
    return TextConfig() & text.data
      .._config(color: _color, background: _bcolor, size: _size)
      .._config(underline: _underline, strike: _strike, overline: _overline)
      .._config(italic: _italic, weight: _weight, shadow: _shadow, ink: _ink)
      // overtap with text
      .._config(color: text._color, background: text._bcolor, size: text._size)
      .._config(underline: text._underline, strike: text._strike, overline: text._overline)
      .._config(italic: text._italic, weight: text._weight, shadow: text._shadow, ink: text._ink);
  }

  // output
  String get text => data;

  String toString() => text;

  TextStyle get compile {
    List<TextDecoration> deco = [];
    if (_underline ?? false) deco.add(TextDecoration.underline);
    if (_strike ?? false) deco.add(TextDecoration.lineThrough);
    if (_overline ?? false) deco.add(TextDecoration.overline);
    return TextStyle(
      color: _color ?? palette.text,
      decoration: TextDecoration.combine(deco),
      fontSize: _size,
      shadows: (_shadow ?? $shadow).compile,
      fontStyle: (_italic ?? false) ? FontStyle.italic : FontStyle.normal,
      fontWeight: _weight,
      backgroundColor: _bcolor,
      textBaseline: TextBaseline.alphabetic,
    );
  }

  GestureRecognizer get recognizer {
    if ((_ink ?? $ink).data == null) return null;
    if (_ink.data.onTap != null) return TapGestureRecognizer()..onTap = () => (_ink.data.onTap ?? () {})();
    if (_ink.data.onLongPress != null) return LongPressGestureRecognizer()..onLongPress = () => (_ink.data.onLongPress ?? () {})();
    return null;
  }

  Widget inline({String data, String fallback = '', List<TextConfig> children}) {
    List<TextSpan> spans = [];
    for (TextConfig config in children) {
      spans.add(TextSpan(
        text: (this | config).data,
        style: (this | config).compile,
        recognizer: (this | config).recognizer,
      ));
    }
    return Container(
      width: (_fill ?? false) ? double.infinity : null,
      child: Text.rich(
          TextSpan(
            text: data ?? this.data ?? fallback,
            style: compile,
            children: spans,
            recognizer: recognizer,
          ), // break
          textAlign: _align,
          maxLines: _maxLines,
          overflow: TextOverflow.ellipsis,
          style: compile),
    );
  }

  Widget highlight({String charseq, String data, String fallback = '', TextConfig config, TextReaction reaction}) {
    List<String> children = data.split(RegExp(charseq));
    List<Match> matches = RegExp(charseq).allMatches(data).toList();
    List<TextConfig> configList = [];
    for (int i = 1; i < children.length; i++) {
      String match = matches[i - 1].group(0);
      configList.add(((this | config) & match).ink(((this | config)._ink ?? $ink) & reaction));
      configList.add(this & children[i]);
    }
    return ((TextConfig() | this) & children.first).inline(data: data, fallback: fallback, children: configList);
  }

  Widget build({String data, String fallback = '', Reaction reaction}) {
    this.data = data ?? this.data ?? fallback;
    final text = Container(
      width: (_fill ?? false) ? double.infinity : null,
      child: Text(this.data, // break
          textAlign: _align,
          maxLines: _maxLines,
          overflow: TextOverflow.ellipsis,
          style: compile),
    );
    return reaction == null ? text : (_ink ?? $ink).build(data: reaction, child: text);
  }
}

class TextReaction extends Reaction {
  final Function(TextConfig text) onTextTap;
  final Function(TextConfig text) onTextLongPress;
  TextReaction({
    Function() onTap,
    Function() onLongPress,
    this.onTextTap,
    this.onTextLongPress,
  }) : super(onTap: onTap, onLongPress: onLongPress);
}
