part of nueli.core.util;

TranxConfig get $tranx => TranxConfig();

/// A class that provides all basic and advanced animation in the library
class TranxConfig extends Config<TranxConfig, Animation<double>, Animation<double>> {
  Animation<double> data;
  Tween<double> _scale;
  Tween<double> _translateX;
  Tween<double> _translateY;
  Tween<double> _height;
  Tween<double> _width;
  Tween<double> _rotation;
  Tween<double> _opacity;

  TranxConfig() {
    data = null;
    _scale = null;
    _translateX = null;
    _translateY = null;
    _height = null;
    _width = null;
    _rotation = null;
  }

  TranxConfig _config({
    Tween<double> scale,
    Tween<double> translateX,
    Tween<double> translateY,
    Tween<double> height,
    Tween<double> width,
    Tween<double> rotation,
    Tween<double> opacity,
  }) {
    _scale = scale ?? _scale;
    _translateX = translateX ?? _translateX;
    _translateY = translateY ?? _translateY;
    _height = height ?? _height;
    _width = width ?? _width;
    _rotation = rotation ?? _rotation;
    _opacity = opacity ?? _opacity;
    return this;
  }

  // input
  TranxConfig pan(Offset begin, Offset end) => pan_horiz(begin.dx, end.dx).pan_vert(begin.dy, end.dy);
  TranxConfig pan_horiz(double begin, double end) => _config(translateX: Tween(begin: begin, end: end)); // ignore: non_constant_identifier_names
  TranxConfig pan_vert(double begin, double end) => _config(translateY: Tween(begin: begin, end: end)); // ignore: non_constant_identifier_names

  TranxConfig scale(double begin, double end, [Alignment alignment = Alignment.center]) => _config(scale: Tween(begin: begin, end: end));

  TranxConfig size(Size begin, Size end, [Alignment alignment = Alignment.center]) => size_horiz(begin.width, end.width).size_vert(begin.height, end.height);
  TranxConfig size_horiz(double begin, double end, [Alignment alignment = Alignment.center]) => _config(width: Tween(begin: begin, end: end)); // ignore: non_constant_identifier_names
  TranxConfig size_vert(double begin, double end, [Alignment alignment = Alignment.center]) => _config(height: Tween(begin: begin, end: end)); // ignore: non_constant_identifier_names
  TranxConfig get fold_horiz => size_horiz(1, 0); // ignore: non_constant_identifier_names
  TranxConfig get fold_vert => size_vert(1, 0); // ignore: non_constant_identifier_names
  TranxConfig get grow_horiz => size_horiz(0, 1); // ignore: non_constant_identifier_names
  TranxConfig get grow_vert => size_vert(0, 1); // ignore: non_constant_identifier_names

  TranxConfig rotate(double begin, double end) => _config(rotation: Tween(begin: begin, end: end));
  TranxConfig fade(double begin, double end) => _config(opacity: Tween(begin: begin, end: end));

  // operation
  @override
  @deprecated
  TranxConfig operator *(double factor) => this;

  @override
  @deprecated
  TranxConfig operator +(TranxConfig operand) => this | operand;

  @override
  @deprecated
  TranxConfig operator -(TranxConfig operand) => this | operand;

  @override
  TranxConfig operator |(TranxConfig tranx) {
    return TranxConfig() & data
      .._config(width: _width, height: _height, opacity: _opacity, rotation: _rotation) // break
      .._config(scale: _scale, translateX: _translateX, translateY: _translateY) // break
      // overlap with element
      .._config(width: tranx._width, height: tranx._height, opacity: tranx._opacity, rotation: tranx._rotation) // break
      .._config(scale: tranx._scale, translateX: tranx._translateX, translateY: tranx._translateY);
  }

  // output
  Animation<double> get compile => data;

  Widget build({Animation<double> data, Animation<double> fallback, Widget child}) {
    this.data = data ?? this.data ?? fallback;
    if (this.data == null) return child;

    if (_scale != null)
      child = ScaleTransition(
        scale: Tween<double>(begin: _scale.begin, end: _scale.end).animate(this.data),
        child: child,
      );

    if (_height != null)
      child = SizeTransition(
        axis: Axis.vertical,
        sizeFactor: Tween<double>(begin: _height.begin, end: _height.end).animate(this.data),
        child: child,
      );

    if (_width != null)
      child = SizeTransition(
        axis: Axis.horizontal,
        sizeFactor: Tween<double>(begin: _width.begin, end: _width.end).animate(this.data),
        child: child,
      );

    if (_translateX != null || _translateY != null) {
      final translateX = _translateX ?? Tween(begin: 0, end: 0);
      final translateY = _translateY ?? Tween(begin: 0, end: 0);
      child = SlideTransition(
        position: Tween<Offset>(
          begin: Offset(translateX.begin, translateY.begin),
          end: Offset(translateX.end, translateY.end),
        ).animate(this.data),
        child: child,
      );
    }

    if (_opacity != null)
      child = FadeTransition(
        opacity: Tween<double>(begin: _opacity.begin, end: _opacity.end).animate(this.data),
        child: child,
      );

    if (_rotation != null)
      child = RotationTransition(
        turns: Tween<double>(begin: _rotation.begin, end: _rotation.end).animate(this.data),
        child: child,
      );
    return child;
  }
}
