library nueli.core.data;

import 'event.dart' show Cycle, CycleReel;
import 'util/--util.dart' show AdvanceList, modulus;

// operations or gate
mixin Subtract<T extends Subtract<T>> {
  T operator -(T operand);
}

mixin Add<T extends Add<T>> {
  T operator +(T operand);
}

mixin Multiply<T extends Multiply<T>> {
  T operator *(T operand);
}

mixin Divide<T extends Divide<T>> {
  T operator /(T operand);
}

mixin AND<T extends AND<T>> {
  T operator &(T operand);
}

mixin OR<T extends OR<T>> {
  T operator |(T element);
}

mixin NOT<T extends NOT<T>> {
  T operator -();
}

/// Reference
/// <span color = '#99aaff'> Begin Ref ~ </span> [$(Object)]<br>
Ref<T> $<T>([T value]) => Ref<T>.construct(value);

class Ref<T> with CycleReel<T> {
  static Map<Link, dynamic> _dataMap = {};
  static Map<Link, Cycle> _cycleMap = {};
  static Link receipt = root;
  final Link link;

  Ref.of(this.link) {
    receipt = link;
    _dataMap.putIfAbsent(link, () => null);
    _cycleMap.putIfAbsent(link, () => Cycle());
  }

  Ref.construct([T value, Cycle<T> cycle]) : link = root / Symbol('${DateTime.now().microsecond}') {
    data = value;
    this.cycle = cycle;
  }

  factory Ref.assign({Link link, T value, Cycle<T> cycle}) {
    return Ref<T>.of(link)
      ..data = value
      ..cycle = cycle;
  }

  T get data => _dataMap.containsKey(link) ? _dataMap[link] ?? fallback : fallback;
  set data(T data) => _dataMap[link] = data;

  destroy() {
    _dataMap.remove(link);
    _cycleMap.remove(link);
  }

  get cycle => _cycleMap.containsKey(link) ? _cycleMap[link] : Cycle();
  set cycle(Cycle<T> cycle) => _cycleMap[link] = cycle;

  String toString() => data.toString();
}

/// <h1>>> Hash Reference </h1>
/// <span color = '#99aaff'> Begin hash ~ </span> [$$({key : data, key : data})]
Hash<T> $$<T>([Map<Link, T> value]) => Hash<T>.construct(value);

class Hash<V> extends Ref<Map<Link, V>> {
  Hash.of(Link link) : super.of(link);
  Hash.construct([Map<Link, V> hash = const {}]) : super.construct(hash);
  factory Hash.assign({Link link, Map<Link, V> value}) => Hash.of(link)..data = value;

  get fallback => {};

  V operator [](Object key) => containsKey(key) ? data[key] : null;
  operator []=(Object key, V value) => data[key] = value;

  bool containsValue(V value) => data.containsValue(value);
  bool containsKey(Link key) => data.containsKey(key);

  List<MapEntry<Link, V>> get entries => data.entries;
  String toString() => data.toString();
}

/// <h1>>> Link </h1>
/// <span color = '#99aaff'> Begin a link ~ </span> [root / #wrapper / #email / #vdt]<br>
/// <span color = '#99aaff'> Default link with no path ~ <span> [root] <br>
const Link root = Link.create([]);

class Link {
  static const nan = Link.create();

  final List<Symbol> _channel;

  const Link.create([this._channel = const []]);

  Symbol get name => _channel.last;
  Link operator /(Symbol t) => Link.create([..._channel, t]);

  Symbol operator [](int offset) {
    if (_channel.isEmpty) return Symbol.empty;
    assert(modulus(offset) + 1 <= _channel.length);
    return (offset > 0) ? _channel[offset] : _channel[_channel.length - 1 + offset];
  }

  operator []=(int offset, Symbol t) {
    if (_channel.isEmpty) return;
    assert(modulus(offset) + 1 <= _channel.length);
    offset > 0 ? _channel[offset] = t : _channel[_channel.length - 1 + offset] = t;
  }

  bool startsWith(Link path) => _channel.startsWith(path._channel);
  bool endsWith(Link path) => this._channel.endsWith(path._channel);

  Ref<T> ref<T>([T data, Cycle<T> cycle]) => Ref.assign(link: this, value: data, cycle: cycle);
  Hash<T> href<T>([Map<Link, T> data]) => Hash.assign(link: this, value: data);

  @override
  String toString() => _channel.join('=>');
}
