part of nueli.ui;

/// <h1>>> Screen </h1>
abstract class Screen extends StatefulWidget with UI {
  final Ref<bool> _main = $(false);

  final Ref<Animation<double>> _animation = $();
  final Ref<Animation<double>> _secondaryAnimation = $();
  final Ref<Animation<double>> _tertiaryAnimation = $();

  final Ref<Duration> _duration = $();
  final Ref<Duration> _secondaryDuration = $();
  final Ref<Duration> _tertiaryDuration = $();

  final Ref<Curve> _curve = $();
  final Ref<Curve> _secondaryCurve = $();
  final Ref<Curve> _tertiaryCurve = $();

  final Ref<TranxConfig> _tranx = $();
  final Ref<TranxConfig> _secondaryTranx = $();
  final Ref<TranxConfig> _tertiaryTranx = $();

  final String name;
  final bool replace;
  final bool closeAllPrev;

  Screen.open(
    BuildContext context, {
    Key key,
    String name,
    // tranx
    TranxConfig primaryTranx,
    TranxConfig secondaryTranx,
    TranxConfig tertiaryTranx,
    // curve
    Curve curve,
    Curve secondaryCurve,
    Curve tertiaryCurve,
    // duration
    Duration duration,
    Duration secondaryDuration,
    Duration tertiaryDuration,
    this.replace = false,
    this.closeAllPrev = false,
  })  : name = '${name ?? DateTime.now().millisecond}/@screen',
        super(key: key) {
    // tranx
    _tranx.data = primaryTranx;
    _secondaryTranx.data = secondaryTranx;
    _tertiaryTranx.data = tertiaryTranx;
    // curve
    _curve.data = curve;
    _secondaryCurve.data = secondaryCurve;
    _tertiaryCurve.data = tertiaryCurve;
    // duration
    _duration.data = duration;
    _secondaryDuration.data = secondaryDuration;
    _tertiaryDuration.data = tertiaryDuration;
    Screens.open(context, screen: this, replace: replace, closeAllPrev: closeAllPrev);
  }

  Screen.build({
    Key key,
    String name,
    TranxConfig primaryTranx,
    TranxConfig secondaryTranx,
    TranxConfig tertiaryTranx,
    this.replace = false,
    this.closeAllPrev = false,
  })  : name = '${name ?? DateTime.now().millisecond}/@screen',
        super(key: key) {
    _tranx.data = primaryTranx ?? $tranx.pan_vert(1, 0).fade(0, 1);
    _secondaryTranx.data = secondaryTranx ?? $tranx;
    _tertiaryTranx.data = tertiaryTranx ?? $tranx.fade(0, 1);
  }

  ScreenRoute get route => ScreenRouteBuilder(screen: this)();

  open(BuildContext context) {
    return Screens.open(context,
        screen: this, // break
        replace: replace,
        closeAllPrev: closeAllPrev);
  }

  @override
  State<Screen> createState();
}

/// Screen state
abstract class ScreenState<T extends Screen> extends ResumableState<T> with TickerProviderStateMixin, UI {
  final GlobalKey<ScaffoldState> _driver = GlobalKey<ScaffoldState>();
  final Ref<Widget> _toast = $();
  final Ref<SplashFactory> _splash = $();
  AnimationController _controller;

  // Main
  Widget get body;
  DecorConfig get scrim => $decor.paint_bg;
  Widget frame(Widget content) => content;

  // Splash
  SplashFactory get splash => null;

  // AppBar
  Widget get appBar => empty;
  DecorConfig get appBarDecor => $decor;

  TranxConfig get appBarTranx => $tranx;
  bool get overlayAppBar => false;
  bool get useStaticAppBar => false;
  double get appBarHeight => hasAppBar ? 64 : 0;
  bool get hasAppBar => !(appBar == empty || appBar == null);

  // FAB
  Widget get fab => null;
  FloatingActionButtonLocation get fabLocation => FloatingActionButtonLocation.endFloat;

  // Drawer
  Widget get drawer => null;
  double get drawerElevation => 8;

  // End Drawer
  Widget get endDrawer => null;
  double get endDrawerElevation => 8;

  // Tranx
  TranxConfig get tranx => $tranx.pan_vert(1, 0).fade(0, 1);
  TranxConfig get secondaryTranx => $tranx;
  TranxConfig get tertiaryTranx => $tranx.fade(0, 1);

  // Curve
  Curve get curve => Curves.easeOutSine;
  Curve get secondaryCurve => curve.flipped;
  Curve get tertiaryCurve => curve;

  // Duration
  Duration get duration => Duration(milliseconds: 800);
  Duration get secondaryDuration => duration;
  Duration get tertiaryDuration => duration;

  bool get delayTertiary => true;

  List<Logic> get logicArray => [];

  FutureOr<bool> onBackPressed() {
    close();
    return false;
  }

  @mustCallSuper
  void init() {}

  @mustCallSuper
  Future<FutureOr<dynamic>> initAsync() async {
    if (logicArray.isNotEmpty) SplashFactory.invalidate();
    await LogicTree.compile(logicArray);
  }

  /// Do not override
  Animation<double> get animation => widget._animation.data ?? AnimationController(vsync: this, value: 1);

  /// Do not override
  Animation<double> get secondaryAnimation => widget._secondaryAnimation.data ?? AnimationController(vsync: this, value: 1);

  /// Do not override
  Animation<double> get tertiaryAnimation => widget._tertiaryAnimation.data ?? _controller;

  bool get hasDrawer => _driver.currentState.hasDrawer;

  bool get hasEndDrawer => _driver.currentState.hasEndDrawer;

  /// Do not override
  void setState(Function fn) {
    _splash.data.setState(() => super.setState(fn));
  }

  /// Do not override
  void openDrawer() => _driver.currentState.openDrawer();

  /// Do not override
  void openEndDrawer() => _driver.currentState.openDrawer();

  /// Do not override
  Widget safeRegion({Widget child, bool withSafeArea = false}) => withSafeArea // break
      ? SafeArea(child: Padding(child: child, padding: EdgeInsets.only(top: appBarHeight)))
      : Padding(child: child, padding: EdgeInsets.only(top: appBarHeight));

  /// Do not override
  close() => Screens.goBack(context);

  /// Do not override
  showToast(Toast toast, [Alignment alignment = const Alignment(0, 0.75)]) => // break
      setState(() => _toast.data = (Align(alignment: alignment, child: toast)));

  @mustCallSuper
  void initState() {
    _splash.data = splash ?? SplashFactory.loader();
    _controller = AnimationController(vsync: this, duration: tertiaryDuration);
    init();
    super.initState();
  }

  @override
  @mustCallSuper
  Widget build(BuildContext context) {
    UI.globalContext = context;

    // Animation
    widget._tertiaryAnimation.data = _controller;

    // Tranx
    widget._tranx.data = widget._tranx.data ?? tranx;
    widget._secondaryTranx.data = widget._secondaryTranx.data ?? secondaryTranx;
    widget._tertiaryTranx.data = widget._tertiaryTranx.data ?? tertiaryTranx;

    // Duration
    widget._duration.data = widget._duration.data ?? duration;
    widget._secondaryDuration.data = widget._secondaryDuration.data ?? secondaryDuration;
    widget._tertiaryDuration.data = widget._tertiaryDuration.data ?? tertiaryDuration;

    // Curve
    widget._curve.data = widget._curve.data ?? curve;
    widget._secondaryCurve.data = widget._secondaryCurve.data ?? secondaryCurve;
    widget._tertiaryCurve.data = widget._tertiaryCurve.data ?? tertiaryCurve;

    WidgetsBinding.instance.addPostFrameCallback((timeStamp) {
      animation.status == AnimationStatus.forward || animation.status == AnimationStatus.reverse
          ? !delayTertiary
              ? _controller.forward()
              : setState(() => _controller.value = 0)
          : _controller.forward();
    });
    animation.status == AnimationStatus.forward || animation.status == AnimationStatus.reverse
        ? !delayTertiary
            ? _controller.forward()
            : setState(() => _controller.value = 0)
        : _controller.forward();
    final plugAnimationIsActive = animation.status == AnimationStatus.forward || animation.status == AnimationStatus.reverse;
    final toastInstance = _toast.data ?? empty;
    final appBarInstance = Container(
      width: double.infinity,
      child: Stack(fit: StackFit.passthrough, children: [
        appBarDecor.size_h(appBarHeight + MediaQuery.of(context).viewPadding.top).stretch_w.build(),
        Material(color: palette.none, child: SafeArea(child: SizedBox(height: appBarHeight, width: double.infinity, child: appBar))),
      ]),
    );
    final drawerInstance = drawer != null && drawer != empty ? Drawer(child: drawer, elevation: drawerElevation) : null;
    final endDrawerInstance = endDrawer != null && endDrawer != empty ? Drawer(child: endDrawer, elevation: endDrawerElevation) : null;
    final fabInstance = fab != null && fab != empty ? $tranx.scale(0, 1).fade(0, 1).build(data: animation, child: fab) : null;
    final bodyInstance = plugAnimationIsActive ? (delayTertiary ? empty : body) : body;
    final frameInstance = Padding(
      padding: EdgeInsets.only(top: overlayAppBar ? 0 : appBarHeight),
      child: frame(tertiaryTranx.build(
          data: CurvedAnimation(parent: tertiaryAnimation, curve: tertiaryCurve, reverseCurve: tertiaryCurve), // break
          child: bodyInstance)),
    );

    return Material(
      color: palette.none,
      child: _splash.data.wrap(
        future: initAsync(),
        child: Stack(children: [
          ConfigTransition(config: ConfigTween($decor, scrim.stretch).animate(animation)),
          WillPopScope(
            onWillPop: () => onBackPressed(),
            child: Stack(children: [
              Scaffold(
                key: _driver,
                floatingActionButton: fabInstance,
                floatingActionButtonLocation: fabLocation,
                backgroundColor: palette.none,
                drawer: drawerInstance,
                endDrawer: endDrawerInstance,
                body: Stack(children: [
                  tranx.build(
                    data: CurvedAnimation(parent: animation, curve: curve, reverseCurve: secondaryCurve),
                    child: secondaryTranx.build(
                      data: CurvedAnimation(parent: secondaryAnimation, curve: secondaryCurve, reverseCurve: curve),
                      child: Stack(children: [frameInstance, useStaticAppBar ? empty : appBarInstance]),
                    ),
                  ),
                  !useStaticAppBar ? empty : appBarInstance,
                  toastInstance,
                ]),
              ),
            ]),
          ),
        ]),
      ),
    );
  }

  @override
  void dispose() {
    _controller.dispose();
    super.dispose();
  }
}

///<h1>>> App </h1>
class App extends StatelessWidget {
  final Screen entry;
  final bool showDebugSemantics;
  final bool showDebugBanner;
  final Function() init;

  App({
    Key key,
    this.entry,
    this.showDebugBanner = false,
    this.showDebugSemantics = false,
    this.init,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    init();
    return MaterialApp(
      debugShowCheckedModeBanner: showDebugBanner,
      showSemanticsDebugger: showDebugSemantics,
      home: entry.._main.data = true,
      color: palette.background,
      theme: _theme,
    );
  }

  static ThemeData get _theme => ThemeData(
        primarySwatch: palette.material, backgroundColor: palette.background, // break
        scaffoldBackgroundColor: palette.background, accentColor: palette.accent, // break
        primaryColor: palette.primary, primaryColorDark: palette.primaryDark,
      );
}

/// Screens
class Screens {
  Screens._();

  static exit() {
    SystemNavigator.pop(animated: true);
  }

  static bool canGoBack(BuildContext context) => Navigator.of(context).canPop();

  static goBack(BuildContext context) {
    if (canGoBack(context)) return Navigator.pop(context);
  }

  static goToPrevModule(BuildContext context) {
    Navigator.of(context).popUntil((route) => route.settings.name.endsWith('/@module'));
  }

  static goHome(BuildContext context) {
    Navigator.of(context).popUntil((route) => route.isFirst);
  }

  static goBackTo(BuildContext context, Route route) {
    Navigator.of(context).popUntil((_route) => _route == route);
  }

  static open(BuildContext context, {Screen screen, bool replace = false, bool closeAllPrev = false}) {
    if (closeAllPrev ?? false) {
      Navigator.pushAndRemoveUntil(context, screen.route, (r) => false);
    } else if (replace ?? false) {
      Navigator.pushReplacement(context, screen.route);
    } else {
      Navigator.push(context, screen.route);
    }
  }
}

/// Screen route builder
class ScreenRouteBuilder {
  Screen screen;

  ScreenRouteBuilder._(this.screen);
  factory ScreenRouteBuilder({@required Screen screen}) => ScreenRouteBuilder._(screen);
  ScreenRoute call() => ScreenRoute(this);
}

/// Screen route
class ScreenRoute extends ModalRoute {
  final ScreenRouteBuilder builder;
  ScreenRoute(this.builder);

  RouteSettings get settings => RouteSettings(name: builder.screen.name);
  Color get barrierColor => palette.none;
  bool get barrierDismissible => false;
  String get barrierLabel => '';

  Widget buildPage(BuildContext context, Animation<double> animation, Animation<double> secondaryAnimation) => builder.screen;

  @override
  Widget buildTransitions(
    BuildContext context,
    Animation<double> animation,
    Animation<double> secondaryAnimation,
    Widget child,
  ) {
    builder.screen._animation.data = animation;
    builder.screen._secondaryAnimation.data = secondaryAnimation;
    return child;
  }

  @override
  bool get maintainState => true;

  @override
  bool get opaque => false;

  @override
  Duration get transitionDuration => builder.screen._duration.data ?? Duration(milliseconds: 350);

  @override
  Duration get reverseTransitionDuration => builder.screen._secondaryDuration.data ?? Duration(milliseconds: 350);
}

/// Toast
class Toast extends Struct with Anim, Stance {
  final Widget Function(BuildContext context) builder;
  final TranxConfig tranx;
  final DecorConfig decor;
  final Duration delay;

  Toast({this.builder, this.decor, TranxConfig tranx, Duration duration = const Duration(seconds: 5)})
      : delay = duration,
        tranx = tranx ?? $tranx.fade(1, 0);

  Toast.text(TextConfig text, this.decor, {TranxConfig tranx, Duration duration = const Duration(seconds: 5)})
      : delay = duration,
        tranx = tranx ?? $tranx.fade(1, 0),
        builder = ((c) => Row(mainAxisSize: MainAxisSize.min, children: [
              $icon.size_16.color(palette.accent).get(CupertinoIcons.info_circle),
              SizedBox(width: 6),
              text.build(fallback: 'OK'),
            ]));

  @override
  Widget get body {
    stance = [builder];
    return Container(
      decoration: decor ?? $decor.paint(palette.background).radii_all(20).stroke(0.5, palette.accent).compile,
      child: builder(context),
      padding: EdgeInsets.symmetric(vertical: 6, horizontal: 8),
    );
  }
}

/// Splash
class SplashFactory extends Struct {
  static bool _isStartup = true;

  final Ref<bool> _isBusy = $(false);
  final Ref<Widget> _child = $();
  final Ref<Future> _futureAction = $();

  final Widget Function(BuildContext) builder;

  SplashFactory(this.builder);

  factory SplashFactory.loader() {
    return SplashFactory((c) => Material(
        color: palette.background,
        child: Center(
          child: CircularProgressIndicator(
            backgroundColor: palette.accent,
            strokeWidth: 2,
          ),
        )));
  }

  static void invalidate() => _isStartup = true;

  @override
  Widget get body {
    if (_child.data == null) return builder(context);
    return Stack(children: [
      !_isStartup
          ? _child.data
          : FutureBuilder(
              builder: (c, s) {
                if (s.connectionState == ConnectionState.none) return _child.data ?? empty;
                if (s.connectionState == ConnectionState.waiting) return builder(context);
                _isStartup = false;
                return _child.data ?? empty;
              },
              future: _futureAction.data),
      _isBusy.data ? Container(color: palette.shade) : empty,
      _isBusy.data ? builder(context) : empty,
    ]);
  }

  Widget wrap({@required Widget child, Future<void> future}) {
    this._child.data = child;
    this._futureAction.data = future;
    return this;
  }
}

///<h1>>> Logic </h1>
abstract class Logic {
  Symbol get channel;
  Algorithm get initializer;
  Algorithm get reviver => algog;
}

///<h1>>> Logic Tree </h1>
class LogicTree {
  static Map<Symbol, dynamic> _channelBus = {};

  static Future<void> compile(List<Logic> logicArray) async {
    WidgetsFlutterBinding.ensureInitialized();
    for (Logic logic in logicArray) {
      final hasInit = !_channelBus.containsKey(logic.channel);
      _channelBus.putIfAbsent(logic.channel, () => 'init');
      hasInit ? await logic.initializer() : await logic.reviver();
    }
  }
}
