part of nueli.ui;

/// <h1>>> Struct </h1>
@immutable
abstract class Struct extends StatefulWidget with UI {
  final Ref<Function(Function() fn)> _setState = $((f) => f());

  Struct({Key key}) : super(key: key);

  Widget get body;

  @mustCallSuper
  Widget populate() => body;

  /// Do not override
  BuildContext get context => UI.globalContext;

  /// Do not override
  Function(Function() fn) setState(Function() fn) => _setState.data(fn);

  StructState<Struct> createState() => StructState();
}

/// Struct state
@protected
class StructState<T extends Struct> extends State<T> with UI {
  @mustCallSuper
  initTools() {
    UI.globalContext = context;
    widget._setState.data = setState;
  }

  @mustCallSuper
  void initState() {
    initTools();
    super.initState();
  }

  @mustCallSuper
  Widget build(BuildContext context) {
    initTools();
    return widget.populate();
  }
}

///<h1>>> Anim </h1>
mixin Anim on Struct {
  final Ref<AnimationController> _controller = $();

  Duration get duration => Duration(milliseconds: 500);
  Duration get reverseDuration => duration;
  Curve get curve => Curves.easeOutExpo;
  Curve get reverseCurve => curve;

  onComplete() {}

  /// Do not override
  Animation<double> get animation => CurvedAnimation(parent: _controller.data, curve: curve, reverseCurve: reverseCurve);

  /// Do not override
  forward({double from}) => _controller.data.forward(from: from);

  /// Do not override
  reverse({double from}) => _controller.data.reverse(from: from);

  /// Do not override
  animateTo(double target, {Duration duration, Curve curve = Curves.linear}) => _controller.data.animateTo(target);

  /// Do not override
  repeat({double min, double max, bool reverse = false, Duration period}) {
    return _controller.data.repeat(min: min, max: max, reverse: reverse, period: period ?? Duration(milliseconds: 500));
  }

  AnimState<Anim> createState() => AnimState();
}

class AnimState<T extends Anim> extends StructState<T> with TickerProviderStateMixin {
  AnimationStatusListener get _status {
    return (s) => s == AnimationStatus.completed || s == AnimationStatus.dismissed ? widget.onComplete() : {};
  }

  initTools() {
    widget._controller.data = widget._controller.data?? AnimationController(
      vsync: this,
      duration: widget.duration,
      reverseDuration: widget.reverseDuration,
    )..addStatusListener(_status);
    super.initTools();
  }

  dispose() {
    widget._controller.data.removeStatusListener(_status);
    widget._controller.data.dispose();
    super.dispose();
  }
}

/// <h1>>> Stance Component </h1>
/// Performs [tranx] on every unique instance by default.<br>
/// ie. whenever [Stance.createState] is called in its lifecycle <br>
/// and [Stance.compose] is different from the former immediate value.
mixin Stance on Anim {
  final Ref<List> _formerStance = $();

  List get firstStance => null;
  List get stance => _formerStance.data;
  void stanceAction() => forward(from: 0);

  /// Do not override
  set stance(List stance) {
    if (_formerStance.data != stance) {
      stanceAction();
      _formerStance.data = stance;
    }
  }

  @override
  StanceState<Stance> createState() => StanceState();
}

class StanceState<T extends Stance> extends AnimState<T> {
  @override
  initState() {
    widget._formerStance.data = widget.firstStance;
    return super.initState();
  }
}
